/*
	IP Query - Displays list of interfaces

	Copyright 2001 Capt. Tara Milana - WolfPack Entertainment
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <gtk/gtk.h>

#include "ipquery_48x48.xpm"
#include "ipquery_32x32.xpm"

#ifndef gbool
# define gbool	guchar
#endif


typedef struct {

	gbool initialized;
	gbool map_state;
	GtkWidget *toplevel, *main_vbox;
	GtkWidget *ifname_combo, *ip_entry, *name_entry;
	GList *ifname_glist;

} win_struct;

#define PROG_NAME		"IPQuery"
#define PROG_VERSION		"0.0.3"

#define CMD_IFCONFIG		"/sbin/ifconfig"
#define HOSTS_FILE		"/etc/hosts"
#define DEF_IFNAME		"ppp0"

#define ISBLANK(c)	(((c) == ' ') || ((c) == '\t'))
#define ISCR(c)		(((c) == '\n') || ((c) == '\r'))


/* Returns a list of interfaces that ifconfig reports. */
static GList *GetInterfacesList(void)
{
	GList *glist = NULL;
#define buf_len         256
        char buf[buf_len];
        FILE *fp;



	fflush(stdout);
	fflush(stderr);
	fp = popen(CMD_IFCONFIG, "r");


        /* Could not run ifconfig? */
        if(fp == NULL)
	    return(glist);

        /* Begin reading process output. */
        while(fgets(buf, buf_len, fp) != NULL)
        {
            buf[buf_len - 1] = '\0';

            /* Check prefix. */
            if(!ISBLANK(*buf) && !ISCR(*buf) && ((*buf) != '\0'))
	    {
		gchar *strptr = g_strdup(buf);
		gchar *strptr2;

		strptr2 = strchr(strptr, '\n');
		if(strptr2 != NULL)
		    (*strptr2) = '\0';

		strptr2 = strchr(strptr, '\t');
                if(strptr2 != NULL)
                    (*strptr2) = '\0';
		strptr2 = strchr(strptr, ' ');
                if(strptr2 != NULL)
                    (*strptr2) = '\0';

		glist = g_list_append(glist, strptr);
	    }
	}

        /* Close process reading. */
        pclose(fp);

#undef buf_len

        return(glist);
}


/* Returns the named address by looking through the HOSTS_FILE and
 * matching up with the given ifaddr string.
 */
static const gchar *QueryNameFromIP(win_struct *win, const gchar *ifaddr)
{
	int ifaddr_len;
#define buf_len         4096
	char buf[buf_len];
#define ifname_len	1024
	static gchar ifname[ifname_len];

	FILE *fp = fopen(HOSTS_FILE, "r");

        /* Could not open hosts file? */
        if(fp == NULL)
        {
            strncpy(ifname, "*cannot open `" HOSTS_FILE "'*", ifname_len);
            ifname[ifname_len - 1] = '\0';
            return(ifname);
        }

        /* Reset ifname return string. */
        strncpy(ifname, "*no match in `" HOSTS_FILE "'*", ifname_len);
        ifname[ifname_len - 1] = '\0';

        /* Get length of interface address string. */
	ifaddr_len = strlen(ifaddr);

        /* Begin reading hosts file. */
        while(fgets(buf, buf_len, fp) != NULL)
        {
            buf[buf_len - 1] = '\0';

            /* Check prefix. */
            if((ifaddr_len > 0) ? !strncmp(buf, ifaddr, ifaddr_len) : 0)
            {
                const gchar *buf_ptr = (const gchar *)buf;
                gchar *strptr;


		/* Seek buf_ptr to start of third argument or highest,
		 * whichever comes first.
		 */
		/* Seek to second. */
                while(!ISBLANK(*buf_ptr) && ((*buf_ptr) != '\0'))
                    buf_ptr++;
		while(ISBLANK(*buf_ptr))
		    buf_ptr++;
                /* Seek to third. */
                while(!ISBLANK(*buf_ptr) && ((*buf_ptr) != '\0'))
                    buf_ptr++;
                while(ISBLANK(*buf_ptr))
                    buf_ptr++;

		/* Copy buf_ptr to ifname. */
		strncpy(ifname, buf_ptr, ifname_len);
		ifname[ifname_len - 1] = '\0';

		strptr = strchr(ifname, ' ');
		if(strptr != NULL)
		    (*strptr) = '\0';
                strptr = strchr(ifname, '\t');
                if(strptr != NULL)
                    (*strptr) = '\0';

		break;
            }
        }

        /* Close hosts file. */
	fclose(fp);

#undef ifname_len
#undef buf_len

        return(ifname);
}

/* Query IP, returns a statically allocated string containing the IP. */
static const gchar *QueryIP(win_struct *win, const gchar *ifname)
{
	int ifname_len, matched_if = 0;
#define buf_len		256
	char buf[buf_len];
#define ifaddr_len	80
	static gchar ifaddr[ifaddr_len];

	FILE *fp;


        fflush(stdout);
        fflush(stderr);
	fp = popen(CMD_IFCONFIG, "r");


	/* Could not run ifconfig? */
	if(fp == NULL)
	{
            strncpy(ifaddr, "*cannot run `" CMD_IFCONFIG "'*", ifaddr_len);
            ifaddr[ifaddr_len - 1] = '\0';
	    return(ifaddr);
	}

	/* Reset ifaddr to "no such interface" error string. */
	strncpy(ifaddr, "*no such interface*", ifaddr_len);
	ifaddr[ifaddr_len - 1] = '\0';

	/* Get length of interface name. */
	ifname_len = strlen(ifname);

	/* Begin reading process output. */
	while(fgets(buf, buf_len, fp) != NULL)
	{
	    buf[buf_len - 1] = '\0';

	    /* Check prefix. */
	    if((ifname_len > 0) ? !strncmp(buf, ifname, ifname_len) : 0)
	    {
		matched_if = 1;	/* Mark that we got matching if. */
	    }
	    else if(matched_if)
	    {
		const gchar *buf_ptr = (const gchar *)buf;
		gchar *strptr;
		const gchar *mstr = "inet addr";

		while(ISBLANK(*buf_ptr))
		    buf_ptr++;

		if(!strncmp(buf_ptr, mstr, strlen(mstr)))
		{
		    while(!isdigit(*buf_ptr) && !ISCR(*buf_ptr) &&
                          ((*buf_ptr) != '\0')
		    )
			buf_ptr++;

		    strncpy(ifaddr, buf_ptr, ifaddr_len);
		    ifaddr[ifaddr_len - 1] = '\0';

		    strptr = strchr(ifaddr, ' ');
		    if(strptr != NULL)
			(*strptr) = '\0';

		    break;
		}
	    }
	}

	/* Close process reading. */
	pclose(fp);

#undef ifaddr_len
#undef buf_len

	return(ifaddr);
}

/* Sets the WM icon for the given GtkWindow widget w. */
static void SetWindowWMIcon(GtkWidget *w, const u_int8_t **icon)
{
	GdkWindow *window;
        GdkPixmap *pixmap;
        GdkBitmap *mask;
        GtkStyle *style;


	window = ((w == NULL) ? NULL : w->window);
	if(window == NULL)
	    return;

	style = gtk_widget_get_style(w);
	if(style == NULL)
	    style = gtk_widget_get_default_style();
	if(style == NULL)
	    return;

        pixmap = gdk_pixmap_create_from_xpm_d(
	    window,
	    &mask,
	    &style->bg[GTK_STATE_NORMAL],
	    (gchar **)icon
        );
        gdk_window_set_icon(window, NULL, pixmap, mask);
/*      gdk_window_set_icon_name(w, "WMHints Test Icon"); */

	return;
}

/* Creates a new icon, returns the fixed widget. */
static GtkWidget *CreateIcon(win_struct *win, const u_int8_t **icon)
{
        GdkWindow *window;
        GtkWidget *w, *pixmap;
        GdkPixmap *gdk_pixmap;
        GdkBitmap *mask;
        GtkStyle *style;
        gint width, height;


	/* Get reference window. */
	w = win->toplevel;
	if(w != NULL)
	{
	    window = w->window;
	    style = gtk_widget_get_style(w);
	    if(style == NULL)
		style = gtk_widget_get_default_style();
	}
	else
	{
	    return(NULL);
	}

	/* Create fixed widget to hold icon pixmap. */
        w = gtk_fixed_new();

	/* Create icon pixmap. */
	gdk_pixmap = gdk_pixmap_create_from_xpm_d(
	    window,
	    &mask,
	    &style->bg[GTK_STATE_NORMAL],
	    (gchar **)icon
	);
	pixmap = gtk_pixmap_new(gdk_pixmap, mask);

	/* Get size of newly loaded pixmap. */
	gdk_window_get_size((GdkWindow *)gdk_pixmap, &width, &height);

        /* Adjust size of fixed widget to fit pixmap. */
        gtk_widget_set_usize(w, width, height);

        /* Put pixmap into fixed widget. */
        if(pixmap != NULL)  
            gtk_fixed_put(GTK_FIXED(w), pixmap, 0, 0);

        /* Unref gdk pixmap so when no other resources are using it, it
         * will be destroyed.
         */   
        gdk_pixmap_unref(gdk_pixmap);

        /* Mask available? */
        if(mask != NULL)
        {
            gtk_widget_shape_combine_mask(w, mask, 0, 0);

            /* Unref bitmap mask so when no other resources are using
             * it, it will be destroyed.
             */
            gdk_bitmap_unref(mask);
        }

        if(pixmap != NULL)
        {
            /* Make pixmap eligable for showing since its a child of the
             * returned fixed widget.
             */
            gtk_widget_show(pixmap);
        }

        return(w);
}               


/* Interface name entry enter callback. */
static void IFNameEnterCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *w;
	const gchar *new_ifname, *new_ifaddr = NULL, *new_name = NULL;
	win_struct *win = (win_struct *)data;
        if(win == NULL)
            return;

	w = widget;
	if(w != NULL)
	{
	    new_ifname = (const gchar *)gtk_entry_get_text(GTK_ENTRY(w));
	    if((new_ifname != NULL) ? ((*new_ifname) != '\0') : 0)
		new_ifaddr = (const gchar *)QueryIP(win, new_ifname);
	}
	w = win->ip_entry;
	if((w != NULL) && (new_ifaddr != NULL))
	{
	    gtk_entry_set_text(GTK_ENTRY(w), new_ifaddr);
	}

	w = win->name_entry;
	if((w != NULL) && (new_ifaddr != NULL))
        {
	    new_name = QueryNameFromIP(win, new_ifaddr);
	    if(new_name != NULL)
		gtk_entry_set_text(GTK_ENTRY(w), new_name);
        }

	return;
}

/* Interface name entry change callback. */
static void IFNameChangeCB(GtkWidget *widget, gpointer data)
{
        gbool in_list = FALSE;
        GtkWidget *w;
	GList *glist;
        const gchar *new_ifname = NULL, *new_ifaddr = NULL, *new_name = NULL;
        win_struct *win = (win_struct *)data;
        if(win == NULL)
            return;

	w = widget;
	if(w != NULL)
	{
	    const gchar *cstrptr;
	    new_ifname = (const gchar *)gtk_entry_get_text(GTK_ENTRY(w));


	    glist = win->ifname_glist;
	    while(glist != NULL)
	    {
		cstrptr = (const gchar *)glist->data;
		if(cstrptr != NULL)
		{
		    if(!strcmp(cstrptr, new_ifname))
		    {
			in_list = TRUE;
			break;
		    }
		}

		glist = glist->next;
	    }
	}

	w = win->ip_entry;
	if((w != NULL) && (new_ifname != NULL) && in_list)
	{
	    new_ifaddr = (const gchar *)QueryIP(win, new_ifname);
	    if(new_ifaddr != NULL)
		gtk_entry_set_text(GTK_ENTRY(w), new_ifaddr);
        }

        w = win->name_entry;
        if((w != NULL) && (new_ifaddr != NULL))
        {
            new_name = QueryNameFromIP(win, new_ifaddr);
            if(new_name != NULL)
                gtk_entry_set_text(GTK_ENTRY(w), new_name);
        }

        return;
}

/* Destroy signal callback. */
static void DestroyCB(GtkObject *object, gpointer data)
{
	return;
}

/* Delete (close) event callback. */
static gint DeleteCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	win_struct *win = (win_struct *)data;
	if(win == NULL)
	    return(FALSE);

	gtk_main_quit();

	return(TRUE);
}

/* Refresh callback. */
static void RefreshCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *w;
	GList *glist, *new_glist;
        win_struct *win = (win_struct *)data;
        if(win == NULL)
            return;

	/* Create new glist. */
	new_glist = GetInterfacesList();
	w = win->ifname_combo;
	if(w != NULL)
	    gtk_combo_set_popdown_strings(GTK_COMBO(w), new_glist);

	/* Delete old glist. */
        glist = win->ifname_glist;
        win->ifname_glist = NULL;
        if(glist != NULL)
        {
            g_list_foreach(glist, (GFunc)g_free, NULL);
            g_list_free(glist);
        }

	/* Record new glist. */
	win->ifname_glist = new_glist;

        return;
}

/* Close callback. */
static void CloseCB(GtkWidget *widget, gpointer data)
{
	DeleteCB(widget, NULL, data);
	return;
}

int main(int argc, char *argv[])
{
	const gchar *def_ifname = DEF_IFNAME;
	const gchar *new_ifaddr, *new_name;
	gchar title_string[512];
	const gchar *cstrptr;
	gint	bw = (80 + (2 * 3)),
		bh = (30 + (2 * 3));
	gint border_major = 5, border_minor = 2;
	GtkWidget *w, *parent, *parent2, *parent3, *parent4;
	GtkCombo *combo;
	GList *glist;
	win_struct *win = (win_struct *)calloc(1, sizeof(win_struct));


	/* Check if any arguments are given. */
	if(argc > 1)
	{
	    cstrptr = (const gchar *)argv[1];
	    if(cstrptr != NULL)
	    {
		/* Help? */
		if(!strcasecmp(cstrptr, "--help") ||
                   !strcasecmp(cstrptr, "-help") ||
                   !strcasecmp(cstrptr, "--h") ||
                   !strcasecmp(cstrptr, "-h") ||
                   !strcasecmp(cstrptr, "-?") ||
                   !strcasecmp(cstrptr, "/?")
		)
		{
		    printf("Usage: %s [interface_name]\n", argv[0]);
		    return(0);
		}
		/* Version? */
		else if(!strcasecmp(cstrptr, "--version") ||
                        !strcasecmp(cstrptr, "-version")
		)
		{
                    printf("%s Version %s\n", PROG_NAME, PROG_VERSION);
		    return(0);
		}
		else
		{
		    def_ifname = cstrptr;
		}
	    }
	}

	gtk_init(&argc, &argv);

	win->initialized = TRUE;
	win->map_state = FALSE;

	win->toplevel = w = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_set_usize(w, 320, -1);
	strcpy(title_string, PROG_NAME);
	strcat(title_string, ": ");
	cstrptr = (const gchar *)getenv("HOSTNAME");
	if(cstrptr == NULL)
	    cstrptr = "*unknown host*";
	strncat(title_string, cstrptr, 256);
	gtk_window_set_title(GTK_WINDOW(w), title_string);
	gtk_signal_connect(
	    GTK_OBJECT(w), "destroy",
	    GTK_SIGNAL_FUNC(DestroyCB),
	    (gpointer)win
	);
        gtk_signal_connect(
            GTK_OBJECT(w), "delete_event",
            GTK_SIGNAL_FUNC(DeleteCB),
            (gpointer)win
        );
	gtk_widget_realize(w);
	SetWindowWMIcon(w, (const u_int8_t **)ipquery_48x48_xpm);
	parent = w;

	win->main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent2 = w;


	/* Hbox to hold icon and entries. */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Icon. */
	w = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, border_major);
	gtk_widget_show(w);
        parent3 = w;

        w = CreateIcon(win, (const u_int8_t **)ipquery_32x32_xpm);
        gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, FALSE, 0);
        gtk_widget_show(w);


	/* Vbox to hold entries. */
	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, border_major);
	gtk_widget_show(w);
        parent3 = w;


	/* Interface name. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
	/* Label. */
	w = gtk_label_new("Interface Name:");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	/* Combo. */
	win->ifname_combo = w = gtk_combo_new();
	combo = GTK_COMBO(w);
        gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	gtk_combo_disable_activate(combo);
	gtk_entry_set_editable(GTK_ENTRY(combo->entry), TRUE);
	gtk_combo_set_case_sensitive(combo, TRUE);
        gtk_signal_connect(
            GTK_OBJECT(combo->entry), "activate",
            GTK_SIGNAL_FUNC(IFNameEnterCB),
            (gpointer)win
        );
        gtk_signal_connect(
            GTK_OBJECT(combo->entry), "changed",
            GTK_SIGNAL_FUNC(IFNameChangeCB),
            (gpointer)win
        );
	gtk_widget_show(w);
	/* Get list of interfaces. */
	win->ifname_glist = glist = GetInterfacesList();
	gtk_combo_set_popdown_strings(combo, glist);
        gtk_entry_set_text(GTK_ENTRY(combo->entry), def_ifname);
	
        /* IP address. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
	/* Label. */
        w = gtk_label_new("IP Address:");
        gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        /* Entry. */
        win->ip_entry = w = gtk_entry_new_with_max_length(80);
        gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
        gtk_entry_set_editable(GTK_ENTRY(w), FALSE);
        gtk_widget_show(w);
	/* Get initial IP. */
	new_ifaddr = (const gchar *)QueryIP(win, def_ifname);
	gtk_entry_set_text(GTK_ENTRY(w), new_ifaddr);

	/* Name address. */
        w = gtk_hbox_new(FALSE, border_minor);
        gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        parent4 = w;
        /* Label. */
        w = gtk_label_new("Name:");
        gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_widget_show(w);
        /* Entry. */
        win->name_entry = w = gtk_entry_new_with_max_length(80);
        gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
        gtk_entry_set_editable(GTK_ENTRY(w), FALSE);
        gtk_widget_show(w);
        /* Get initial IP. */
        new_name = QueryNameFromIP(win, new_ifaddr);
        gtk_entry_set_text(GTK_ENTRY(w), new_name);


	/* Separator. */
	parent = win->main_vbox;
	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Buttons hbox. */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Buttons. */
	w = gtk_button_new_with_label("Refresh");
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
	gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(RefreshCB),
            (gpointer)win
        );
	gtk_widget_show(w);

        w = gtk_button_new_with_label("Close");
        gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT | GTK_HAS_DEFAULT);
        gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(CloseCB),    
            (gpointer)win
        );
        gtk_widget_show(w);



	/* Map toplevel. */
	w = win->toplevel;
	if(w != NULL)
	{
	    gtk_widget_show(w);
	    win->map_state = TRUE;
	}

	gtk_main();


#define DO_DESTROY_WIDGET	\
{ \
 if(w != NULL) \
  gtk_widget_destroy(w); \
}

	w = win->ifname_combo;
	win->ifname_combo = NULL;
	DO_DESTROY_WIDGET

	glist = win->ifname_glist;
	win->ifname_glist = NULL;
	if(glist != NULL)
	{
	    g_list_foreach(glist, (GFunc)g_free, NULL);
	    g_list_free(glist);
	}

	w = win->toplevel;
	win->toplevel = NULL;
	DO_DESTROY_WIDGET;

#undef DO_DESTROY_WIDGET

	g_free(win);
	win = NULL;

	return(0);
}
